// Copyright (c) 2005 Alexander S. Ovcharenko, SymbioForge.
//
//  FntV1to2 - utility which allows to build font binary resource version 2
// from font resources version 1.
//
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// FREE SOFTWARE FUNDATION; either version 2 of the License or (at your
// opinion) any later version.
//
// To obtain a complete GNU License from internet use URL http://www.fsf.org
//                                                                           
//  -------------------------------------------------------------------------
//    To compile this file use the following command line options:
//  gcc fntv1to2.c toollib.c -o fntv1to2.c
//
//    Where GCC is a GNU C compiler or any other suitable to your computer 
// operation system;

// Raster palm OS font resource version 2 have four internal glyphs images
// parts of different density:
// Single, One and Half, Double, Triple, Quadruple density;
// The sizes of all font are based on single density (base) part of font
// resource.

#include <stdio.h>
#include "toollib.h"

#define maxIndexOfDensities	5	// Maximum number of density allowed to use.
#define maxNumberOfResources	32	// Maximum number of resources to use.
#define baseFontBitsInRow	16	// Number of bits in one row element in the base font.

// Local structutre that keeps command line arguments;
typedef struct 
	{
	 char			name[ filePathLength+1 ]; // Name of binary resource file to build;
	 UInt16			multiplyValue;		// Used to enlarge or reduce glyph images.
	 UInt8			version;     		// Font resource version;
	 UInt16			densityValue; 		// Density mode to use this resource to.
	 Boolean		saved;       		// Flag when resource already saved.
	} fontResType;

typedef struct {
	 UInt16			widthTable[ maxTableElements ];
	 UInt16			bitmapLocationTable[ maxTableElements ];
	 UInt16			height;				// Corrected glyphs height.
	 UInt32			rowBytes;			// Number bytes per row;
	 UInt32			glyphsImageWidthInPixels;	// Summary of all images width;
	 UInt8			headerLength;			// Length of header to write.
	 UInt8			densityPartLength;		// Length of density part. For PalmOS v5 or PalmOS v6.
	} correctedFontValuesType;

typedef struct {
	 UInt16			resourceSequence;	// Sequence number of resource used in the command line.
	 char			nameOfFile[ filePathLength+1 ];
	 UInt32			length;      		// Length of a resource;
	 FontType		fontV1Header;
	 UInt16			bitmapLocationTable[ maxTableElements ];
	 UInt16			offsetWidthTable[ maxTableElements ];
	 UInt8			version;		// Version 1.
	 FontDensityType	densityType;
	 UInt16			percentToScale;		// Percent value to scale from base image to image of font dimensions;
	 UInt16			percentToDownScale;	// suragateFontPart[ index ].percentToDownScale = 25; // Percent value to scale to single density.
	 UInt16			bitsInRow;		// Number of bits of one row element.
	 correctedFontValuesType corrected;		// 
       } SuragateFontPartType;

// A local structure with program arguments data.
struct {
	Boolean			helpScreen;
	Boolean			badArgument; // Set true if argument in command line was bad.
	Boolean			headerFormatV5; // Set true if header format is for PalmOS version 5.x;
	UInt16			baseResourceIndex;
 	char			nameOfResourceFile[ filePathLength+1 ];
       } argsData;

// A header for fontV2type font.
FontTypeV2			fontV2Header;		// Header of version type 2 font.
correctedFontValuesType		baseFontV2values;	// Corrected values of FontV2.
SuragateFontPartType		suragateFontPart[ maxIndexOfDensities ];

UInt16				numberOfFontParts = 0;   // Total number of font resources used.
UInt16				mostLowDensityIndex = 0; 
FILE				* fileOutPtr=NULL, * tmpFilePtr=NULL, * resFilePtr=NULL;
UInt8				originalGlyphData[ bufferLength ];
glyphBufferType			* originalGlyphPtr = NULL;
unsigned long			fileSize;

UInt8				dotsToShiftLeft, dotsToShiftRight;
UInt8				minGlyphWidth=0, minGlyphWidthInBytes=0, minGlyphHeight=0;
UInt16				rowCounter, columnCounter;
UInt8				glyphLeftByte, glyphRightByte;
UInt8				shiftRowWidth, shiftRowWidthInBytes;
UInt8				tileBytes[ maxGlyphColumnHeight ];
Boolean				wasWarningMessage;

static void usage( void );	// Display help screen.

main( int argc, char * argv[] )
 {
   short			i, i1, passNumber, arglen;
   short			resNameLength=0; 
   UInt32			uniqID;
   FILE				* outFilePtr=NULL, * inFilePtr=NULL;
   Boolean			fontResourceDefined=false;
   UInt16			tablesLengthInWords, glyphWidth=0;
   UInt32			fntOffsetWidthTblPos, fntGlyphBitmapPos, fntBitmapLocationPos;
   char				*densityMsgPtr[5];

   // Set text variables for density messages.
   densityMsgPtr[0]="single";
   densityMsgPtr[1]="one and a half";
   densityMsgPtr[2]="double";
   densityMsgPtr[3]="triple";
   densityMsgPtr[4]="quadruple";

   puts("FntV1To2 - Build PalmOS(tm) nfnt font resource version 2 from fonts version 1." );
   puts("Version 1.3 - (c) 2005, SymbioForge (http://www.symbioforge.com/feii/)." );

   fillAreaByByte( &argsData, 0, sizeof(argsData) );
   fillAreaByByte( &suragateFontPart[ 0 ], 0, sizeof( SuragateFontPartType ) * maxIndexOfDensities );

   argsData.headerFormatV5=true;
   baseFontV2values.headerLength=fontTag2V5Length;
   baseFontV2values.densityPartLength=fontDensityV5Length;

   for( i=1; i < argc; i++ )
    {
     arglen = strlen( argv[i] ); // Get length of this argument;
     if ( ( arglen >= 1 ) && (argv[i][0]=='-' ||  argv[i][0]=='+' || argv[i][0]=='/') )
        {   // Check for help request.
 	    if ( (arglen==2) && ( tolower(argv[ i ][ 1 ])=='h' || argv[ i ][ 1 ]=='?' ) )
		argsData.helpScreen=1;
	    else  // Check for density operations argument set;
	    if( (arglen==3) && (tolower(argv[ i ][ 1 ])=='v') )
               {
		if ( tolower(argv[ i ][ 2 ])=='5' )
		   {
			argsData.headerFormatV5=true;
			baseFontV2values.densityPartLength=fontDensityV5Length;
			baseFontV2values.headerLength=fontTag2V5Length;
		   }
		 else
		if ( tolower(argv[ i ][ 2 ])=='6' )
		   {
			argsData.headerFormatV5=false;
			baseFontV2values.densityPartLength=fontDensityV6Length;
			baseFontV2values.headerLength=fontTag2Length;
		   }
		else
		   {
     		 	puts( "\nError: unsupported font header format. Program stopped.\00" );
  		 	errorExit( errorExitValue );

		   };
	       }
	    else
	    if ( arglen>=3 )
		{ char		firstChar, secondChar;
	 	  char	*	inFontBinaryNamePtr = NULL; // Binary font resource name for input;
		  UInt16	densityValue=0, index=0;

		  firstChar = tolower( argv[ i ][ 1 ] );
		  secondChar = tolower( argv[ i ][ 2 ] );

		  // Extract file name of font for input.
		  inFontBinaryNamePtr = arglen > 3 ? &argv[ i ][ 3 ] : NULL;
		  resNameLength=strlen( inFontBinaryNamePtr );

		  // Sort command line resources.
		  // Single Density goes firts, One and Half second and etc...
		  if ( (firstChar=='s' || firstChar=='1' || firstChar=='b' ) && (secondChar=='d') ) // Single Density.
		   {
			index=0;
			suragateFontPart[ index ].densityType.density = singleDensityValue;
			suragateFontPart[ index ].percentToScale = 100;
			suragateFontPart[ index ].percentToDownScale = 100;
			suragateFontPart[ index ].bitsInRow = 16;
		   }
		  else
		  if ( (firstChar=='o') && (secondChar=='h' || secondChar=='d' ) ) // One and a Half density.
		   {
			index=1;
			suragateFontPart[ index ].densityType.density = oneAndAHalfDensityValue;
			suragateFontPart[ index ].percentToScale = 150;
			// Percent value to scale to single density.
			suragateFontPart[ index ].percentToDownScale = 66;
			suragateFontPart[ index ].bitsInRow = 24;
//			suragateFontPart[ index ].bitsInRow = 32;
		    }
                  else
                  if ( (firstChar=='d' || firstChar=='2' ) && (secondChar=='d') ) // Double density.
		   {
			index=2;
			suragateFontPart[ index ].densityType.density = doubleDensityValue;
			suragateFontPart[ index ].percentToScale = 200;
			// Percent value to scale to single density.
			suragateFontPart[ index ].percentToDownScale = 50;
			suragateFontPart[ index ].bitsInRow = 32;
		   }
                  else
		  if ( (firstChar=='t' || firstChar=='3' ) && (secondChar=='d') ) // Triple density.
		   {
			index=3;
			suragateFontPart[ index ].densityType.density = tripleDensityValue;
			suragateFontPart[ index ].percentToScale = 300;
			suragateFontPart[ index ].percentToDownScale = 33; // Percent value to scale to single density.
			suragateFontPart[ index ].bitsInRow = 48;
		   }
		  else
                  if ( (firstChar=='q' || firstChar=='4' ) && (secondChar=='d') ) // Quadruple density.
		   {
			index=4;
			suragateFontPart[ index ].densityType.density = quadrupleDensityValue;
			suragateFontPart[ index ].percentToScale = 400;
			suragateFontPart[ index ].percentToDownScale = 25; // Percent value to scale to single density.
			suragateFontPart[ index ].bitsInRow = 64;
		   }
                  else
		   {
     		 	puts( "\nError: invalid argument used. Program stopped.\00" );
  		 	errorExit( errorExitValue );
		   };

	  	  if ( ( resNameLength > 0) && (inFontBinaryNamePtr!=NULL) ) // Set resource file name.
		    {
			if ( resNameLength > filePathLength )
			   {
     		 		puts( "\nError: font resource name is too long. Program stopped.\00" );
  		 		errorExit( errorExitValue );
			   };

			// Check if resource name already defined.
			if ( (suragateFontPart[ index ].nameOfFile==NULL) || (strlen( suragateFontPart[ index ].nameOfFile)==0) )
			 {
		           strcpy( suragateFontPart[ index ].nameOfFile, inFontBinaryNamePtr );
			   suragateFontPart[ index ].resourceSequence = numberOfFontParts;

			   if ( numberOfFontParts==0 )
				argsData.baseResourceIndex = index; // Remind what density goes first in the command line arguments;

		           numberOfFontParts++;
			 }
		 	else
			 {
			   printf( "\nError: resource of %s density part already defined. Program stopped.\n\00", densityMsgPtr[ index ] );
  		 	   errorExit( errorExitValue );
			 };
		    }
		  else
		     {
			printf( "\nError: file name of %s density not specified. Program stopped.\00", densityMsgPtr[ index ] );
  		 	errorExit( errorExitValue );
		     };
                }
	       else
	        {
     		 	puts( "\nError: unsupported argument used. Program aborted.\00" );
  		 	errorExit( errorExitValue );
		};
         } // When no flag '-' or '+' assume this is a resource file name to build.
        else        
	 {  char	* outFontBinaryNamePtr=NULL; // Binary font resource name to build;

	    outFontBinaryNamePtr = &argv[ i ][ 0 ];
	    resNameLength = strlen( outFontBinaryNamePtr );
	    if ( resNameLength > filePathLength )
		{
     		 puts( "\nError: font resource name is too long. Program stopped.\00" );
  		 errorExit( errorExitValue );
		};
	    strcpy( argsData.nameOfResourceFile, outFontBinaryNamePtr);
        };
   }; // End of argument counting;

if ( strlen( argsData.nameOfResourceFile )==0 && numberOfFontParts==0 )
	argsData.helpScreen = true;
else
if ( strlen( argsData.nameOfResourceFile )==0 )
    {
	puts("\nError: font file name for build is not set. Program stopped.\00");
	errorExit( errorExitValue );
    }
else
if ( numberOfFontParts==0 )
    {
	puts( "\nError: there is no resource to add in the argument line. Program stopped.\00" );
	errorExit( errorExitValue );
    }
else // Treat font resource;
    {
      Boolean	noFontResourceSet	= true;

      // Build some versionType1 header elements.
      fillAreaByByte( &fontV2Header, 0, fontTag2V6Length );
      fontV2Header.fontType		= 0x9000 | fntExtendedFormatMask;
      fontV2Header.version		= 1;
      fontV2Header.densityCount		= numberOfFontParts;
      fontV2Header.reserved		= 0;

      // Check every resource for version, load its header and original tables;
      for( i=0; i < maxIndexOfDensities; i++ )
	  {
	      if ( suragateFontPart[ i ].densityType.density !=0 )
		{ 
		  FILE 		* chkFilePtr=NULL;
		  UInt8		fontVersion;

		  // Check version of font resource file;
	  	  chkFilePtr = openFile( suragateFontPart[ i ].nameOfFile, "rb" );
		  suragateFontPart[ i ].length = ftell( chkFilePtr );
	          // Check font resource for consistency.
	          suragateFontPart[ i ].version = getFontResourceType( chkFilePtr, true );

	          if ( suragateFontPart[ i ].version==0 )
		    {
	  	     printf("\nError: file '%s' is not a binary font resource. Program stopped.\n\00", suragateFontPart[ i ].nameOfFile );
	  	     errorExit( errorFileOperation );
		    }
     	          else
	          if ( suragateFontPart[ i ].version!=1 )
		    {
	  	     printf("\nError: font resource '%s' is not type version 1. Program aborted.\n\00", suragateFontPart[ i ].nameOfFile );
	  	     errorExit( errorFileOperation );
		    };

		  // Get font header.
		  getFontHeaderV1( chkFilePtr, &suragateFontPart[ i ].fontV1Header );

		  // Get resource table length;
		  tablesLengthInWords = ( suragateFontPart[ i ].fontV1Header.lastGlyph - suragateFontPart[ i ].fontV1Header.firstGlyph )+3;
		  (UInt32)fntBitmapLocationPos = (suragateFontPart[ i ].fontV1Header.owTLoc-tablesLengthInWords-5 )*2L+(UInt32)fontTag1Length;

		  // Use owtLocate instead this.
		  (UInt32)fntOffsetWidthTblPos = (UInt32)fntBitmapLocationPos+(UInt32)tablesLengthInWords * 2;

		  fileSeek( chkFilePtr, (UInt32)fntBitmapLocationPos );
		  for ( i1=0; i1< tablesLengthInWords; i1++ )
		      suragateFontPart[ i ].bitmapLocationTable[ i1 ] = (UInt16)fileGetWord( chkFilePtr );
		  for ( i1=0; i1< (tablesLengthInWords-1); i1++ ) // Do not read padded word;
		      suragateFontPart[ i ].offsetWidthTable[ i1 ] = (UInt16)fileGetWord( chkFilePtr );

		  suragateFontPart[ i ].offsetWidthTable[ tablesLengthInWords-1 ] = 0;

		  fclose( chkFilePtr );
		  if ( noFontResourceSet ) // noFontResourceSet flag;
		    {
		  	noFontResourceSet = false;
			mostLowDensityIndex = i;
		    };
		};
	  };

      if ( noFontResourceSet )
	 {
     	   puts( "\nError: at least one font resource file have to be set. Program stopped.\00" );
	   errorExit( errorExitValue );
	 };
    };

if ( argsData.helpScreen ) // Display program help.
 {
  usage();
  return;
 };

printf("\nResource to build: '%s'\n", argsData.nameOfResourceFile );

// *** Continue to generate font version type 2 header.
fontV2Header.lastGlyph = suragateFontPart[ argsData.baseResourceIndex ].fontV1Header.lastGlyph;
fontV2Header.firstGlyph = suragateFontPart[ argsData.baseResourceIndex ].fontV1Header.firstGlyph;
tablesLengthInWords = ( fontV2Header.lastGlyph - fontV2Header.firstGlyph )+3; // Length in words.
fontV2Header.owTLoc = ( ( baseFontV2values.headerLength + fontV2Header.densityCount * baseFontV2values.densityPartLength+tablesLengthInWords*2l )/2)-8; // Offset to width table;

fontV2Header.fRectHeight = scaleBackValueByPercent(  suragateFontPart[ argsData.baseResourceIndex ].fontV1Header.fRectHeight,
		      				 suragateFontPart[ argsData.baseResourceIndex ].percentToScale,
		      				 useCeilingWhenScaleHeight );

fontV2Header.maxWidth = scaleBackValueByPercent( suragateFontPart[ argsData.baseResourceIndex ].fontV1Header.maxWidth, 
		      				 suragateFontPart[ argsData.baseResourceIndex ].percentToScale,
		      				 useCeilingWhenScaleWidth );

fontV2Header.kernMax = scaleBackValueByPercent( suragateFontPart[ argsData.baseResourceIndex ].fontV1Header.kernMax,
		      				 suragateFontPart[ argsData.baseResourceIndex ].percentToScale,
		      				 useCeilingWhenScaleWidth );

fontV2Header.fRectWidth = scaleBackValueByPercent( suragateFontPart[ argsData.baseResourceIndex ].fontV1Header.fRectWidth,
		      				 suragateFontPart[ argsData.baseResourceIndex ].percentToScale,
		      				 useCeilingWhenScaleWidth );

fontV2Header.ascent = scaleBackValueByPercent( suragateFontPart[ argsData.baseResourceIndex ].fontV1Header.ascent, 
		      				 suragateFontPart[ argsData.baseResourceIndex ].percentToScale,
		      				 useCeilingWhenScaleHeight );

fontV2Header.descent = scaleBackValueByPercent( suragateFontPart[ argsData.baseResourceIndex ].fontV1Header.descent,
		      				 suragateFontPart[ argsData.baseResourceIndex ].percentToScale,
		      				 useCeilingWhenScaleHeight );

fontV2Header.leading = scaleBackValueByPercent( suragateFontPart[ argsData.baseResourceIndex ].fontV1Header.leading,
		      				 suragateFontPart[ argsData.baseResourceIndex ].percentToScale,
		      				 useCeilingWhenScaleHeight );

fontV2Header.nDescent = 0;
baseFontV2values.height = fontV2Header.fRectHeight;
baseFontV2values.rowBytes = 0;
baseFontV2values.glyphsImageWidthInPixels = 0;
baseFontV2values.bitmapLocationTable[ 0 ] = 0;

// Scaleback nearest font density for single density tables.
for ( i1=0; i1 < (tablesLengthInWords-1); i1++ )
  {
    baseFontV2values.widthTable[ i1 ] = ( suragateFontPart[ argsData.baseResourceIndex ].bitmapLocationTable[ i1+1 ] - suragateFontPart[ argsData.baseResourceIndex ].bitmapLocationTable[ i1 ] ) & 0xFF;
    baseFontV2values.widthTable[ i1 ] = scaleBackValueByPercent( baseFontV2values.widthTable[ i1 ], suragateFontPart[ argsData.baseResourceIndex ].percentToScale, useCeilingWhenScaleWidth );
    baseFontV2values.bitmapLocationTable[ i1+1 ] = baseFontV2values.bitmapLocationTable[ i1 ] + baseFontV2values.widthTable[ i1 ];
    baseFontV2values.glyphsImageWidthInPixels += baseFontV2values.widthTable[ i1 ];
    baseFontV2values.widthTable[ i1 ]|= (baseFontV2values.widthTable[i1]==0) ? 0xFFFF : 0;
  };

fontV2Header.rowWords=(baseFontV2values.glyphsImageWidthInPixels+baseFontBitsInRow-1)/baseFontBitsInRow;
baseFontV2values.rowBytes = fontV2Header.rowWords * (baseFontBitsInRow/8);

// Create location tables for every density in the font. 
for( i=0; i < maxIndexOfDensities; i++ )
  if ( suragateFontPart[ i ].densityType.density !=0 )
	{
  		suragateFontPart[ i ].corrected.glyphsImageWidthInPixels = 0;
		suragateFontPart[ i ].corrected.bitmapLocationTable[ 0 ] = 0;

  		for ( i1=0; i1< tablesLengthInWords-1; i1++ )
     			{
			  glyphWidth = (baseFontV2values.widthTable[i1]&0xFF)!=0xFF ? (baseFontV2values.widthTable[ i1 ]&0xFF) : 0;
			  glyphWidth = scaleValueByPercent( glyphWidth, suragateFontPart[i].percentToScale, useCeilingWhenScaleWidth );

			  suragateFontPart[i].corrected.widthTable[i1]=glyphWidth;
  			  suragateFontPart[i].corrected.glyphsImageWidthInPixels+=glyphWidth;
			  suragateFontPart[i].corrected.bitmapLocationTable[i1]=scaleValueByPercent( baseFontV2values.bitmapLocationTable[i1], suragateFontPart[i].percentToScale, useCeilingWhenScaleWidth );
			};

		suragateFontPart[i].corrected.widthTable[tablesLengthInWords-1] = 0;

		suragateFontPart[i].corrected.rowBytes = (scaleValueByPercent( baseFontV2values.rowBytes, suragateFontPart[i].percentToScale, useCeilingWhenScaleWidth )+1)&~1;
		suragateFontPart[i].corrected.height = scaleValueByPercent( fontV2Header.fRectHeight, suragateFontPart[i].percentToScale, useCeilingWhenScaleHeight );
	};

// Offset in bytes from the begining of the font data to the start of the font image for this density;
(UInt32)fntGlyphBitmapPos = (UInt32)baseFontV2values.headerLength+
			    (UInt32)baseFontV2values.densityPartLength*
			    (UInt32)fontV2Header.densityCount+
			    (UInt32)(tablesLengthInWords * 4l);

// Buld extended density headers.
for(  i=0;  i < maxIndexOfDensities; i++ )
  if ( suragateFontPart[  i  ].densityType.density !=0  )
	{
	  suragateFontPart[ i ].densityType.glyphBitsOffset = (UInt32)fntGlyphBitmapPos;
	  (UInt32)fntGlyphBitmapPos+= (UInt32)suragateFontPart[ i ].corrected.rowBytes*(UInt32)suragateFontPart[i].corrected.height;
	};

// build font version 2 in temporary file.
tmpFilePtr = tmpfile();
if ( tmpFilePtr==NULL )
 {
    printf("\nError: can not open temporary file. Program aborted.\n\00");
    errorExit( errorFileOperation );
 };

printf("Write font header information in %s format...", argsData.headerFormatV5 ? "PalmOS(tm) v5.x" : "Cobalt(tm)" ); // Write font version Type2 header to file.
if( argsData.headerFormatV5 )
  putFontHeaderV2osV5( &fontV2Header, tmpFilePtr ); // Write font header PalmOS(tm) 5.x format.
else
  putFontHeaderV2( &fontV2Header, tmpFilePtr ); // Write font header for Cobalt(tm) format.

for( i=0; i < maxIndexOfDensities; i++ ) // Write density parts.
 if ( suragateFontPart[ i ].densityType.density !=0 )
    if( argsData.headerFormatV5 )
	putFontDensityPartV5( &suragateFontPart[ i ].densityType, tmpFilePtr );
    else
	putFontDensityPart( &suragateFontPart[ i ].densityType, tmpFilePtr );
puts(" Done.");

printf("Write glyphs information tables..."); // Add tables to file.
for ( i=0; i < tablesLengthInWords; i++ )
	filePutWord( tmpFilePtr, baseFontV2values.bitmapLocationTable[ i ] );
for ( i=0; i < tablesLengthInWords; i++ )
	filePutWord( tmpFilePtr, baseFontV2values.widthTable[ i ] );

// Initialyze glyph images areas for every density part.
originalGlyphPtr = (glyphBufferType *)&originalGlyphData[ 0 ];
fillAreaByByte( originalGlyphPtr, 0xDB, bufferLength );
for( i=0; i < maxIndexOfDensities; i++ )
if ( suragateFontPart[ i ].densityType.density !=0 )
  {
    for( i1=0; i1 < suragateFontPart[ i ].corrected.height; i1++ )
	if ( fwrite(originalGlyphPtr, 1,(UInt32)suragateFontPart[i].corrected.rowBytes, tmpFilePtr)!=(UInt32)suragateFontPart[i].corrected.rowBytes )
		{
			puts("\nError: write to file operation failed. Program aborted.\00");
		  	errorExit( errorFileOperation );
		};
  };
printf(" Done.");

// Now read resource files and write font images to temporary file as different density parts.
// Needs some optimization.
for( i = 0; i < maxIndexOfDensities; i++ )
 if ( suragateFontPart[ i ].densityType.density !=0 )
    { UInt16 glyphNo;

       resFilePtr = openFile( suragateFontPart[ i ].nameOfFile, "rb" );
       printf("\nBuild %s density font part from '%s'...", densityMsgPtr[ i ], suragateFontPart[ i ].nameOfFile );

       // Check for corrected glyph height;
       if ( suragateFontPart[ i ].fontV1Header.fRectHeight != suragateFontPart[ i ].corrected.height )
	 {
	   printf( "\n* Warning: incorrect font height %u pixels, required %u.\00", suragateFontPart[ i ].fontV1Header.fRectHeight, suragateFontPart[ i ].corrected.height );
	   wasWarningMessage = true;
	 };

        //  Load glyph image to the buffer.
       for( i1=0; i1 < (tablesLengthInWords-1) ; i1++ )
	{
	  fillAreaByByte( originalGlyphPtr, 0, bufferLength );
	  glyphNo = suragateFontPart[i].fontV1Header.firstGlyph+i1;

	  originalGlyphPtr->width = suragateFontPart[ i ].bitmapLocationTable[ i1+1 ]-suragateFontPart[ i ].bitmapLocationTable[ i1 ];
	  originalGlyphPtr->widthInBytes = (originalGlyphPtr->width+7)>>3;
          originalGlyphPtr->height = suragateFontPart[ i ].fontV1Header.fRectHeight;
	  originalGlyphPtr->bufferPtr = (UInt8 *)originalGlyphPtr+sizeof(glyphBufferType)+1;

	  if ( (originalGlyphPtr->width!=0) &&
	      ( ((suragateFontPart[ i ].corrected.widthTable[ i1 ]&0xFF)==0xFF) ||  ((suragateFontPart[ i ].corrected.widthTable[ i1 ]&0xFF)==0 )) )
	      {
	        printf( "\n* Warning: image of glyph (ASCII: %u) skiped for default.\00", glyphNo );
                      wasWarningMessage = true;
	      };

	  if ( originalGlyphPtr->width==0 )
	   {
	     if ( ((suragateFontPart[ i ].corrected.widthTable[ i1 ]&0xFF)!=0xFF) &&  ((suragateFontPart[ i ].corrected.widthTable[ i1 ]&0xFF)!=0x0) )
	      {
	          printf( "\n* Warning: image for glyph (ASCII: %u) is absent.\00", glyphNo );
                        wasWarningMessage = true;
	       };
	      continue;
	   };

	  minGlyphWidth=min( originalGlyphPtr->width, suragateFontPart[i].corrected.widthTable[ i1 ]&0xFF );
          minGlyphHeight = min( originalGlyphPtr->height, suragateFontPart[ i ].corrected.height );
	  minGlyphWidthInBytes=( minGlyphWidth+7 )>>3;

	  dotsToShiftLeft = suragateFontPart[ i ].bitmapLocationTable[ i1 ] & 7; // How many bits to shift right when add first byte.
	  dotsToShiftRight = ( 8-dotsToShiftLeft ) & 7; // How many bits to shift left at adding next byte;

	  // Load glyph image to the buffer;
	  for( rowCounter=0; rowCounter < originalGlyphPtr->height; rowCounter++ )
	   for( columnCounter=0; columnCounter < originalGlyphPtr->widthInBytes; columnCounter++ )
		{
		  // Get original byte of glyph from resource.
		  fileSeek( resFilePtr, (UInt32)fontTag1Length+
					(UInt32)(suragateFontPart[i].bitmapLocationTable[i1]>>3)+
					(UInt32)columnCounter+
					(UInt32)rowCounter * (UInt32)suragateFontPart[ i ].fontV1Header.rowWords * 2l );
		  glyphLeftByte = fileGetByteWFName( resFilePtr, suragateFontPart[ i ].nameOfFile );
		  glyphRightByte = dotsToShiftLeft!=0 ? fileGetByteWFName( resFilePtr, suragateFontPart[ i ].nameOfFile ) : 0;
		  originalGlyphPtr->bufferPtr[ (rowCounter * (UInt32)originalGlyphPtr->widthInBytes)+columnCounter ] |=
			( (glyphLeftByte<<dotsToShiftLeft)&0xFF) | ( (glyphRightByte>>dotsToShiftRight)&0xFF );
		};

	  // Warn when glyph image width is incorrected;
          if ( originalGlyphPtr->width != ( suragateFontPart[ i ].corrected.widthTable[ i1 ]&0xFF ) )
	   {
	    if (glyphNo <= suragateFontPart[ i ].fontV1Header.lastGlyph)
	       printf( "\n* Warning: incorrect (ASCII: %u) glyph width - %u pixels (required %u).\00", glyphNo, originalGlyphPtr->width, suragateFontPart[ i ].corrected.widthTable[ i1 ]&0xFF );
	    else
	       printf( "\n* Warning: incorrect (default) glyph width - %u pixels (required %u).\00", originalGlyphPtr->width, suragateFontPart[ i ].corrected.widthTable[ i1 ]&0xFF );
            wasWarningMessage = true;
	   };

/*
	 // Display every glyph.
	 printf("Glyph no: %u\n", i1+suragateFontPart[ i ].fontV1Header.firstGlyph );
	 for( rowCounter=0; rowCounter < minGlyphHeight; rowCounter++ )
	   {
	    for( columnCounter=0; columnCounter < originalGlyphPtr->width; columnCounter++ )
		 printf( "%s", ( ( originalGlyphPtr->bufferPtr[ (columnCounter>>3)+(rowCounter*originalGlyphPtr->widthInBytes) ] & ( 0x80>>( columnCounter & 0x07 ) ) ) !=0 ) ? "#" : "-" );
	    puts("");
	   };
     	 puts("");
*/

	 // Shift glyph right and add last bits of previous glyph.
	 if ( (suragateFontPart[ i ].corrected.widthTable[ i1 ]!=0xFFFF) && (originalGlyphPtr->width!=0) )
	   {
	    // Check if we need to shif glyph right.
            if ( (suragateFontPart[ i ].corrected.bitmapLocationTable[ i1 ]&0x07)!=0 )
	      {	
		UInt8	shiftRow[ rowLength+1 ];

		// Add last part of previous glyph which does not fit to byte;
		dotsToShiftRight=suragateFontPart[i].corrected.bitmapLocationTable[i1]&0x07;
		dotsToShiftLeft=(8-dotsToShiftRight) & 0x07; // How many bits to shift left at adding next byte;
 	 	shiftRowWidth=minGlyphWidth+dotsToShiftRight;
		shiftRowWidthInBytes=(shiftRowWidth+7) >> 3;

		for( rowCounter=0; rowCounter < minGlyphHeight; rowCounter++ )
		  {
			for( columnCounter = 0; columnCounter < shiftRowWidthInBytes; columnCounter++ )
			  {
			   if ( columnCounter==0 )
				shiftRow[ columnCounter ] = ( tileBytes[rowCounter] & (0xFF << dotsToShiftLeft) ) | ( originalGlyphPtr->bufferPtr[rowCounter * originalGlyphPtr->widthInBytes] >> dotsToShiftRight );
			   else
			   	shiftRow[ columnCounter ] = ( originalGlyphPtr->bufferPtr[ columnCounter-1 + (rowCounter * originalGlyphPtr->widthInBytes) ] << dotsToShiftLeft ) |
							    ( ( columnCounter < originalGlyphPtr->widthInBytes ) ?
		                               		      ( originalGlyphPtr->bufferPtr[ columnCounter + (rowCounter * originalGlyphPtr->widthInBytes) ] >> dotsToShiftRight ) : 0 );
			  };

	      		proceedBinaryDataFileOut( tmpFilePtr, &shiftRow[0],
							(UInt32)suragateFontPart[i].densityType.glyphBitsOffset+
							(UInt32)(suragateFontPart[i].corrected.bitmapLocationTable[i1]>>3)+
							suragateFontPart[i].corrected.rowBytes*(UInt32)rowCounter,
            			        		shiftRowWidthInBytes );

			// Accumulate last bytes of shifted row to use in the next glyph write;
	      		tileBytes[ rowCounter ] = shiftRow[ (shiftRowWidthInBytes-1) ];
		  };
	      }
           else // Write glyph to the temporary file;
	     for( rowCounter=0; rowCounter < minGlyphHeight; rowCounter++ )
	      {
	        proceedBinaryDataFileOut( tmpFilePtr, &originalGlyphPtr->bufferPtr[ rowCounter * originalGlyphPtr->widthInBytes ],
					  (UInt32)suragateFontPart[i].densityType.glyphBitsOffset+
					  (UInt32)(suragateFontPart[i].corrected.bitmapLocationTable[i1]>>3)+
					  suragateFontPart[i].corrected.rowBytes*(UInt32)rowCounter,
            			          originalGlyphPtr->widthInBytes );
	        tileBytes[ rowCounter ] = originalGlyphPtr->bufferPtr[ (originalGlyphPtr->widthInBytes-1) + rowCounter * originalGlyphPtr->widthInBytes ];
	      };
        }; // End of glyph width check condition;
    }; // End of glyph counter block;


 fclose( resFilePtr );
 printf("%sDone.", wasWarningMessage ? "\n" : " " );
 wasWarningMessage = false;
}; // End of density counting block;

// Move file from temporary file to a binary font resource;
fileOutPtr = openFile( argsData.nameOfResourceFile, "wb+" );
if ( fileOutPtr == NULL )
 {
   printf("\nError: can not write to file '%s' for write. Program aborted.\n\00", argsData.nameOfResourceFile );
   errorExit( errorFileOperation );
 };

 (unsigned long)fileSize = (unsigned long)fileLength( tmpFilePtr );
 proceedBinaryDataFileTransfer( tmpFilePtr, fileOutPtr, 0l, (unsigned long)fileSize );

 fclose( fileOutPtr );
 fclose( tmpFilePtr );
};

// Display user's help.
static void usage( void )
{
  puts("\nUsage ([] means optional):");
  puts("   fntv1to2 [flags] <fntV2res.bin> +<density><fntV1res.bin>");
  puts("            ... where fntV2res.bin is a name of font resource to build.");
  puts("                      fntV1res.bin is a font resource version 1 file(s).\n");
  puts("Density are:\n");
  puts("+SD,+1D,+BD        single (base) density font part.");
  puts("+OH,+OD            one and a half density font part.");
  puts("+DD,+2D            double density font part.");
  puts("+TD,+3D            triple density font part.");
  puts("+QD,+4D            quadruple density font part.\n");
  puts("Flags are:\n");
  puts("-v5	   build font with PalmOS(tm) version 5.x header format. (Default).");
  puts("-v6	   build font with modified Cobalt(tm) version 6.x header format.");
  puts("-h, -?	   this help screen.");
};
